_MINANGLE = -190.0
_MAXANGLE = 190.0


class MyCobotDataException(Exception):
    pass


def check_boolean(b):
    if b != 0 and b != 1:
        raise MyCobotDataException("This parameter needs to be 0 or 1")


def check_range(v, ra):
    min, max = ra
    if min <= v <= max:
        return True
    else:
        return False


def check_coord(id, v):
    coords = ["x", "y", "z", "rx", "ry", "rz"]
    if id < 3:
        pass
    else:
        if not check_range(v, [-180, 180]):
            raise MyCobotDataException(
                "{} value not right, should be -180 ~ 180".format(coords[id])
            )


def check_rgb(args):
    rgb_str = ["r", "g", "b"]
    for i, v in enumerate(args):
        if not check_range(v, [0, 255]):
            raise MyCobotDataException(
                "The RGB value needs be 0 ~ 255, but the %s is %s" % (rgb_str[i], v)
            )


def check_datas(**kwargs):
    if kwargs.get("joint_id", None) is not None and not 0 < kwargs["joint_id"] < 7:
        raise MyCobotDataException(
            "The joint id not right, should be 1 ~ 6, but received %s."
            % kwargs["joint_id"]
        )
    if (
        kwargs.get("degree", None) is not None
        and not _MINANGLE <= kwargs["degree"] <= _MAXANGLE
    ):
        raise MyCobotDataException(
            "degree value not right, should be -180 ~ 180, but received %s"
            % kwargs["degree"]
        )
    if kwargs.get("len6", None) is not None:
        len_ = len(kwargs["len6"])
        if len_ != 6:
            raise MyCobotDataException(
                "The length of data should be 6, the truth is %s" % len_
            )
    if kwargs.get("degrees", None) is not None:
        for idx, angle in enumerate(kwargs["degrees"]):
            if not _MINANGLE <= angle <= _MAXANGLE:
                raise MyCobotDataException(
                    "degree value not right, should be -180 ~ 180, the error index is %s"
                    % idx
                )
    if kwargs.get("speed", None) is not None and not 0 <= kwargs["speed"] <= 100:
        raise MyCobotDataException(
            "speed value not right, should be 0 ~ 100, the error speed is %s"
            % kwargs["speed"]
        )
    if kwargs.get("rgb", None) is not None:
        check_rgb(kwargs["rgb"])



import time
import math
import serial
import logging

from pymycobot.log import setup_logging
from pymycobot.generate import MycobotCommandGenerater
from pymycobot.common import Command
from pymycobot.error import check_datas


class MyCobot(MycobotCommandGenerater):
    """MyCobot Python API Serial communication class.

    Supported methods:

        # Overall status
            Look at parent class: `MycobotCommandGenerater`.

        # MDI mode and operation
            get_radians()
            send_radians()
            sync_send_angles() *
            sync_send_coords() *
            Other look at parent class: `MycobotCommandGenerater`.

        # JOG mode and operation
            Look at parent class: `MycobotCommandGenerater`.

        # Running status and Settings
            Look at parent class: `MycobotCommandGenerater`.

        # Servo control
            Look at parent class: `MycobotCommandGenerater`.

        # Atom IO
            Look at parent class: `MycobotCommandGenerater`.

        # Basic
            Look at parent class: `MycobotCommandGenerater`.

        # Other
            wait() *
    """

    def __init__(self, port, baudrate="115200", timeout=0.1, debug=False):
        """
        Args:
            port     : port string
            baudrate : baud rate string, default '115200'
            timeout  : default 0.1
            debug    : whether show debug info
        """
        super(MyCobot, self).__init__(debug)
        self.debug = debug
        setup_logging(self.debug)
        self.log = logging.getLogger(__name__)
        self._serial_port = serial.Serial(port, baudrate, timeout=timeout)

    def _write(self, command):
        self.log.debug("_write: {}".format(command))

        self._serial_port.write(command)
        self._serial_port.flush()
        time.sleep(0.05)

    def _read(self):
        if self._serial_port.inWaiting() > 0:
            data = self._serial_port.read(self._serial_port.inWaiting())
            self.log.debug("_read: {}".format(data))
        else:
            self.log.debug("_read: no data can be read")
            data = None
        return data

    def _mesg(self, genre, *args, **kwargs):
        """

        Args:
            genre: command type (Command)
            *args: other data.
                   It is converted to octal by default.
                   If the data needs to be encapsulated into hexadecimal,
                   the array is used to include them. (Data cannot be nested)
            **kwargs: support `has_reply`
                has_reply: Whether there is a return value to accept.
        """
        real_command, has_reply = super(MyCobot, self)._mesg(genre, *args, **kwargs)
        self._write(self._flatten(real_command))

        if has_reply:
            data = self._read()
            res = self._process_received(data, genre)
            if genre in [
                Command.IS_POWER_ON,
                Command.IS_CONTROLLER_CONNECTED,
                Command.IS_PAUSED,
                Command.IS_IN_POSITION,
                Command.IS_MOVING,
                Command.IS_SERVO_ENABLE,
                Command.IS_ALL_SERVO_ENABLE,
                Command.GET_SERVO_DATA,
                Command.GET_DIGITAL_INPUT,
                Command.GET_GRIPPER_VALUE,
                Command.IS_GRIPPER_MOVING,
                Command.GET_SPEED,
            ]:
                return self._process_single(res)
            elif genre in [Command.GET_ANGLES]:
                return [self._int_to_angle(angle) for angle in res]
            elif genre in [Command.GET_COORDS]:
                if res:
                    r = []
                    for idx in range(3):
                        r.append(self._int_to_coord(res[idx]))
                    for idx in range(3, 6):
                        r.append(self._int_to_angle(res[idx]))
                    return r
                else:
                    return res
            elif genre in [Command.GET_JOINT_MIN_ANGLE, Command.GET_JOINT_MAX_ANGLE]:
                return self._int_to_angle(res[0]) if res else 0
            else:
                return res
        return None

    def get_radians(self):
        """Get all angle return a list

        Return:
            data_list (list[radian...]):
        """
        angles = self._mesg(Command.GET_ANGLES, has_reply=True)
        return [round(angle * (math.pi / 180), 3) for angle in angles]

    def send_radians(self, radians, speed):
        """Send all angles

        Args:
            radians (list): example [0, 0, 0, 0, 0, 0]
            speed (int): 0 ~ 100
        """
        check_datas(len6=radians, speed=speed)
        degrees = [self._angle_to_int(radian * (180 / math.pi)) for radian in radians]
        return self._mesg(Command.SEND_ANGLES, degrees, speed)

    def sync_send_angles(self, degrees, speed, timeout=7):
        t = time.time()
        self.send_angles(degrees, speed)
        while time.time() - t < timeout:
            f = self.is_in_position(degrees, 0)
            if f:
                break
            time.sleep(0.1)
        return self

    def sync_send_coords(self, coords, speed, mode, timeout=7):
        t = time.time()
        self.send_coords(coords, speed, mode)
        while time.time() - t < timeout:
            if self.is_in_position(coords, 1):
                break
            time.sleep(0.1)
        return self

    # Basic for raspberry pi.
    def gpio_init(self):
        """Init GPIO module.
        Raspberry Pi version need this.
        """
        import RPi.GPIO as GPIO

        GPIO.setmode(GPIO.BCM)
        self.gpio = GPIO

    def gpio_output(self, pin, v):
        """Set GPIO output value.
        Args:
            pin: port number(int).
            v: Output value(int), 1 - GPIO.HEIGH, 2 - GPIO.LOW
        """
        self.gpio.setup(pin, self.gpio.OUT)
        self.gpio.setup(pin, v)

    # Other
    def wait(self, t):
        time.sleep(t)
        return self



import re
import serial.tools.list_ports


def get_port_list():
    plist = [str(x).split(" - ")[0].strip() for x in serial.tools.list_ports.comports()]
    return plist


def detect_port_of_basic():
    """Detect M5 Basic.
    Returns the serial port string of the first detected M5 Basic.
    If it is not found, it returns `None`.
    """
    type_re = re.compile(r"PID=[0-9a-zA-Z]+:([0-9a-zA-Z]+)")
    plist = [x for x in serial.tools.list_ports.comports()]
    for port in plist:
        res = type_re.findall(port[2])
        if res and res[0].lower() == "ea60":
            return port[0]
    else:
        return None



import os
import sys

PYTHON_VERSION = sys.version_info[:2]
if (2, 7) != PYTHON_VERSION < (3, 5):
    print("This mycobot version requires Python2.7, 3.5 or later.")
    sys.exit(1)

import setuptools

os.environ["_MYCOBOT_VERSION_ONLY"] = "1"
import pymycobot

if PYTHON_VERSION == (2, 7):
    long_description = ""
else:
    long_description = open("README.md").read() + open("docs/README.md").read()

setuptools.setup(
    name="pymycobot",
    version=pymycobot.__version__,
    author=pymycobot.__author__,
    author_email=pymycobot.__email__,
    description="Python API for serial communication of MyCobot.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=pymycobot.__git_url__,
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["pyserial"],
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, != 3.4.*",
)



import logging
import sys

from pymycobot.log import setup_logging
from pymycobot.error import check_datas
from pymycobot.common import Command, DataProcessor


class MycobotCommandGenerater(DataProcessor):
    """MyCobot Python API (Chain operation: *)

    Supported methods:

        # Overall status
            version()
            power_on()
            power_off()
            is_power_on()
            release_all_servos()
            is_controller_connected()

        # MDI mode and operation
            get_angles()
            send_angle()
            send_angles()
            get_coords()
            send_coords()
            is_in_position()
            is_moving() x

        # JOG mode and operation
            jog_angle()
            jog_coord()
            jog_stop()
            set_encoder()
            get_encoder()
            set_encoders()
            pause()
            resume()
            stop()
            is_paused()

        # Running status and Settings
            get_encoder()
            set_encoder()
            get_encoders()
            set_encoders()
            get_speed()
            set_speed() *
            get_joint_min_angle()
            get_joint_max_angle()

        # Servo control
            is_servo_enable()
            is_all_servo_enable()
            set_servo_data()
            get_servo_data()
            set_servo_calibration()
            release_servo()
            focus_servo()

        # Atom IO
            set_color() *
            set_led_color()
            set_pin_mode()
            set_digital_output()
            get_digital_input()
            set_pwm_mode() x
            set_pwm_output()
            get_gripper_value()
            set_gripper_state() *
            set_gripper_value()
            set_gripper_ini()
            is_gripper_moving()

        # Basic
            set_basic_output()
            get_basic_input()
    """

    def __init__(self, debug=False):
        """
        Args:
            port     : port string
            baudrate : baud rate string, default '115200'
            timeout  : default 0.1
            debug    : whether show debug info
        """
        self._version = sys.version_info[:2][0]
        self.debug = debug
        setup_logging(self.debug)
        self.log = logging.getLogger(__name__)

    def _mesg(self, genre, *args, **kwargs):
        """

        Args:
            genre: command type (Command)
            *args: other data.
                   It is converted to octal by default.
                   If the data needs to be encapsulated into hexadecimal,
                   the array is used to include them. (Data cannot be nested)
            **kwargs: support `has_reply`
                has_reply: Whether there is a return value to accept.
        """
        command_data = self._process_data_command(args)

        LEN = len(command_data) + 2
        command = [
            Command.HEADER,
            Command.HEADER,
            LEN,
            genre,
            command_data,
            Command.FOOTER,
        ]

        real_command = self._flatten(command)
        has_reply = kwargs.get("has_reply", False)

        return real_command, has_reply

    # System status
    def version(self):  # TODO: test method <11-03-21, yourname> #
        """Get cobot version

        Return:
            mycobot   : 1
            mycobotPro: 101
        """
        recv = self._mesg(Command.VERSION, has_reply=True)
        return recv

    # Overall status
    def power_on(self):
        """Open communication with Atom."""
        return self._mesg(Command.POWER_ON)

    def power_off(self):
        """Close communication with Atom."""
        return self._mesg(Command.POWER_OFF)

    def is_power_on(self):
        """Adjust robot arm status

        Return:
            1 : power on
            0 : power off
            -1: error data
        """
        return self._mesg(Command.IS_POWER_ON, has_reply=True)

    def release_all_servos(self):
        return self._mesg(Command.RELEASE_ALL_SERVOS)

    def is_controller_connected(self):
        """Wether connected with Atom."""
        return self._mesg(Command.IS_CONTROLLER_CONNECTED, has_reply=True)

    """
    def set_free_mode(self, flag):  # TODO:no finish
        if flag:
            self._mesg(Command.SET_FREE_MODE, 1)
        else:
            self._mesg(Command.SET_FREE_MODE, 0)

    def is_free_mode(self):  # TODO: no finish
        return self._process_single(self._mesg(Command.IS_FREE_MODE, has_reply=True))
    """

    # MDI mode and operation
    def get_angles(self):
        """Get all angle return a list.

        Return:
            data_list (list[angle...]):
        """
        return self._mesg(Command.GET_ANGLES, has_reply=True)

    def send_angle(self, id, degree, speed):
        """Send one angle.

        Args:
            id (common.Angle/int): Joint number.
            degree (float): angle value.
            speed (int): 0 ~ 100
        """
        check_datas(joint_id=id, degree=degree, speed=speed)
        return self._mesg(Command.SEND_ANGLE, id, [self._angle_to_int(degree)], speed)

    # @check_parameters(Command.SEND_ANGLES)
    def send_angles(self, degrees, speed):
        """Send all angles

        Args:
            degrees (list): example [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
            speed (int): 0 ~ 100
        """
        check_datas(degrees=degrees, len6=degrees, speed=speed)
        degrees = [self._angle_to_int(degree) for degree in degrees]
        # data = [degrees, speed]
        return self._mesg(Command.SEND_ANGLES, degrees, speed)

    def get_coords(self):
        """Get all coords.

        Return:
            data_list (list): [x, y, z, rx, ry, rz]
        """
        return self._mesg(Command.GET_COORDS, has_reply=True)

    def send_coord(self, id, coord, speed):
        """Send one position of coordination.

        Args:
            id(common.Coord/int): coordination number.
            coord(float): mm
            speed(int): 0 ~ 100
        """
        check_datas(speed=speed)
        return self._mesg(Command.SEND_COORD, id, [self._coord_to_int(coord)], speed)

    def send_coords(self, coords, speed, mode):
        """Send all coordinations.

        Args:
            coords: [x(mm), y, z, rx(angle), ry, rz]
            speed(int);
            mode(int): 0 - normal, 1 - angluar, 2 - linear
        """
        check_datas(len6=coords, speed=speed)
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord_to_int(coords[idx]))
        for idx in range(3, 6):
            coord_list.append(self._angle_to_int(coords[idx]))
        return self._mesg(Command.SEND_COORDS, coord_list, speed, mode)

    def is_in_position(self, data, id):
        """Determine whether the given position is reached.

        Args:
            id: 1 - coords, 0 - angles
            data: The value list of angles or coordinations.

        Return:
            0 : error position
            1 : right position
            -1: error data
        """
        check_datas(len6=data)
        if id == 1:
            data_list = []
            for idx in range(3):
                data_list.append(self._coord_to_int(data[idx]))
            for idx in range(3, 6):
                data_list.append(self._angle_to_int(data[idx]))
        elif id == 0:
            data_list = [self._angle_to_int(i) for i in data]
        else:
            raise Exception("id is not right, please input 0 or 1")

        return self._mesg(Command.IS_IN_POSITION, data_list, id, has_reply=True)

    def is_moving(self):
        """

        Return:
            0 : not moving
            1 : is moving
            -1: error data
        """
        return self._mesg(Command.IS_MOVING, has_reply=True)

    # JOG mode and operation
    def jog_angle(self, joint_id, direction, speed):
        """Control joint in JOG mode.

        Args:
            joint_id(int): Joint number.
            direction: int [0, 1]
            speed: int (0 - 100)
        """
        return self._mesg(Command.JOG_ANGLE, joint_id, direction, speed)

    def jog_coord(self, coord_id, direction, speed):
        """Control coordination in JOG mode.

        Args:
            coord: string
            direction: int [0, 1]
            speed: int (0 - 100)
        """
        return self._mesg(Command.JOG_COORD, coord_id, direction, speed)

    def jog_stop(self):
        """Stop JOG movement."""
        return self._mesg(Command.JOG_STOP)

    def pause(self):
        return self._mesg(Command.PAUSE)

    def is_paused(self):
        return self._mesg(Command.IS_PAUSED, has_reply=True)

    def resume(self):
        return self._mesg(Command.RESUME)

    def stop(self):
        return self._mesg(Command.STOP)

    def set_encoder(self, joint_id, encoder):
        """Set joint encoder value.

        Args:
            joint_id: Joint id 1 - 7
            encoder: The value of the set encoder.
        """
        return self._mesg(Command.SET_ENCODER, joint_id, [encoder])

    def get_encoder(self, joint_id):
        return self._mesg(Command.GET_ENCODER, joint_id, has_reply=True)

    def set_encoders(self, encoders, sp):
        return self._mesg(Command.SET_ENCODERS, encoders, sp)

    def set_encoders(self):
        return self._mesg(Command.GET_ENCODERS, has_reply=True)

    # Running status and Settings
    def get_speed(self):
        return self._mesg(Command.GET_SPEED, has_reply=True)

    def set_speed(self, speed):
        """Set speed value

        Args:
            speed (int): 0 - 100
        """
        check_datas(speed=speed)
        return self._mesg(Command.SET_SPEED, speed)

    """
    def get_feed_override(self):
        return self._process_single(
            self._mesg(Command.GET_FEED_OVERRIDE, has_reply=True)
        )

    def get_acceleration(self):
        return self._process_single(
            self._mesg(Command.GET_ACCELERATION, has_reply=True)
        )
    """

    def get_joint_min_angle(self, joint_id):
        check_datas(joint_id=joint_id)
        return self._mesg(Command.GET_JOINT_MIN_ANGLE, joint_id, has_reply=True)

    def get_joint_max_angle(self, joint_id):
        check_datas(joint_id=joint_id)
        return self._mesg(Command.GET_JOINT_MAX_ANGLE, joint_id, has_reply=True)

    # Servo control
    def is_servo_enable(self, servo_id):
        return self._mesg(Command.IS_SERVO_ENABLE, servo_id)

    def is_all_servo_enable(self):
        return self._mesg(Command.IS_ALL_SERVO_ENABLE, has_reply=True)

    def set_servo_data(self, servo_no, data_id, value):
        return self._mesg(Command.SET_SERVO_DATA, servo_no, data_id, value)

    def get_servo_data(self, servo_no, data_id):
        return self._mesg(Command.GET_SERVO_DATA, servo_no, data_id, has_reply=True)

    def set_servo_calibration(self, servo_no):
        return self._mesg(Command.SET_SERVO_CALIBRATION, servo_no)

    def release_servo(self, servo_id):
        """Power off designated servo

        Args:
            servo_id: 1 ~ 6
        """
        return self._mesg(Command.RELEASE_SERVO, servo_id)

    def focus_servo(self, servo_id):
        """Power on designated servo

        Args:
            servo_id: 1 ~ 6

        """
        return self._mesg(Command.FOCUS_SERVO, servo_id)

    # Atom IO
    def set_color(self, r=0, g=0, b=0):
        """Set the light color

        Args:
            r (int): 0 ~ 255
            g (int): 0 ~ 255
            b (int): 0 ~ 255

        """
        check_datas(rgb=[r, g, b])
        return self._mesg(Command.SET_COLOR, r, g, b)

    def set_pin_mode(self, pin_no, pin_mode):
        """Set the state mode of the specified pin in atom.

        Args:
            pin_no   (int):
            pin_mode (int): 0 - input, 1 - output, 2 - input_pullup
        """
        return self._mesg(Command.SET_PIN_MODE, pin_no, pin_mode)

    def set_digital_output(self, pin_no, pin_signal):
        """

        Args:
            pin_no     (int):
            pin_signal (int): 0 / 1
        """
        return self._mesg(Command.SET_DIGITAL_OUTPUT, pin_no, pin_signal)

    def get_digital_input(self, pin_no):
        return self._mesg(Command.GET_DIGITAL_INPUT, pin_no, has_reply=True)

    """
    def set_pwm_mode(self, pin_no, channel):
        self._mesg(Command.SET_PWM_MODE, pin_no, channel)
    """

    def set_pwm_output(self, channel, frequency, pin_val):
        return self._mesg(Command.SET_PWM_OUTPUT, channel, [frequency], pin_val)

    def get_gripper_value(self):
        """Get the value of gripper."""
        return self._mesg(Command.GET_GRIPPER_VALUE, has_reply=True)

    def set_gripper_state(self, flag, speed):
        """Set gripper switch

        Args:
            flag  (int): 0 - open, 1 - close
            speed (int): 0 ~ 100
        """
        return self._mesg(Command.SET_GRIPPER_STATE, flag, speed)

    def set_gripper_value(self, value, speed):
        """Set gripper value

        Args:
            value (int): 0 ~ 4096
            speed (int): 0 ~ 100
        """
        check_datas(speed=speed)
        return self._mesg(Command.SET_GRIPPER_VALUE, [value], speed)

    def set_gripper_ini(self):
        """Set the current position to zero

        Current position value is `2048`.
        """
        return self._mesg(Command.SET_GRIPPER_INI)

    def is_gripper_moving(self):
        """Judge whether the gripper is moving or not

        Returns:
            0 : not moving
            1 : is moving
            -1: error data
        """
        return self._mesg(Command.IS_GRIPPER_MOVING, has_reply=True)

    # Basic
    def set_basic_output(self, pin_no, pin_signal):
        """Set basic output for M5 version.

        Args:
            pin_no: pin port number.
            pin_signal: 0 / 1
        """
        return self._mesg(Command.SET_BASIC_OUTPUT, pin_no, pin_signal)

    def get_basic_input(self, pin_no):
        """Get basic input for M5 version.

        Args:
            pin_no: pin port number.
        """
        return self._mesg(Command.GET_BASIC_OUTPUT, pin_no, has_reply=True)



from __future__ import absolute_import
import os

if os.environ.get("_MYCOBOT_VERSION_ONLY") != "1":
    from pymycobot.mycobot import MyCobot
    from pymycobot.generate import MycobotCommandGenerater
    from pymycobot.genre import Angle, Coord

    __all__ = ["MyCobot", "MycobotCommandGenerater", "Angle", "Coord", "utils"]

__version__ = "2.5.6"
__author__ = "Zachary zhang"
__email__ = "lijun.zhang@elephantrobotics.com"
__git_url__ = "https://github.com/elephantrobotics/pymycobot"

PI_PORT = "/dev/ttyAMA0"
PI_BAUD = 1000000
